import numpy as np
from scipy.spatial.distance import cdist

from ..base import Base
from ..utils.distance import euclidean, haversine


class SpatialAverage(Base):
    """
    Class to interpolate by fitting a XGBoost Regressor to given
    data.
    Note that radius you specify must be in kilometres if you are passing latitude and longitude as inputs
    """

    def __init__(
        self, radius=100, resolution="standard", coordinate_type="Euclidean", **kwargs
    ):
        super().__init__(resolution, coordinate_type)
        self.radius = radius
        if self.coordinate_type == 'Geographic':
            self.distance = haversine
        elif self.coordinate_type == 'Euclidean':
            self.distance = euclidean
        else:
            raise NotImplementedError(
                "Only Geographic and Euclidean Coordinates are available")

    def _fit(self, X, y):
        """Function for fitting.
        This function is not supposed to be called directly.
        """
        self.X = X
        self.y = y
        return self

    def _predict_grid(self, x1lim, x2lim):
        """Function for grid interpolation.
        This function is not supposed to be called directly.
        """
        # getting the boundaries for interpolation
        x1min, x1max = x1lim
        x2min, x2max = x2lim

        # building the grid
        x1 = np.linspace(x1min, x1max, self.resolution)
        x2 = np.linspace(x2min, x2max, self.resolution)
        X1, X2 = np.meshgrid(x1, x2)
        return self._predict(np.asarray([X1.ravel(), X2.ravel()]).T)

    def _predict(self, X):
        """Function for interpolation on specific points.
        This function is not supposed to be called directly.
        """
        return self._average(X)

    def _average(self, X):
        dist = self.distance(X, self.X)
        mask = self.radius >= dist
        return (self.y*mask).sum(axis=1)/mask.sum(axis=1)



"""
This is a module for Kriging Interpolation
"""
import numpy as np 
from ..base import Base
from pykrige.ok import OrdinaryKriging
from pykrige.uk import UniversalKriging


class Kriging(Base):
    """A class that is declared for performing Kriging interpolation.
    Kriging interpolation (usually) works on the principle of finding the 
    best unbiased predictor. Ordinary Kriging, for an example, involves finding out the
    best unbaised linear predictor. 

    Parameters
    ----------
    type : str, optional
    This parameter defines the type of Kriging under consideration. This 
    implementation uses PyKrige package  (https://github.com/bsmurphy/PyKrige).
    The user needs to choose between "Ordinary" and "Universal".

    plotting: boolean, optional
    This parameter plots the fit semivariogram. We use PyKrige's inbuilt plotter for the same.s
    
    variogram_model : str, optional
    Specifies which variogram model to use; may be one of the following:
    linear, power, gaussian, spherical, exponential, hole-effect.
    Default is linear variogram model. To utilize a custom variogram model,
    specify 'custom'; you must also provide variogram_parameters and
    variogram_function. Note that the hole-effect model is only technically
    correct for one-dimensional problems.

    require_variance : Boolean, optional
    This variable returns the uncertainity in the interpolated values using Kriging
    interpolation. If this is True, kindly call the attribute return_variance, of this class
    to retreive the computed variances. False is the default value.d
    
	nlags: int, optional
	Number of lags to be considered for semivariogram. As in PyKrige, we set default to be 6.
    """

    def __init__(
        self,
        type = "Ordinary",
        plotting = False,
        variogram_model = 'linear',
        require_variance = False,
        resolution = "standard",
        coordinate_type = "Eucledian",
        nlags = 6
    ):

        super().__init__(resolution, coordinate_type)
        self.variogram_model = variogram_model
        self.ok = None
        self.uk = None
        self.type = type 
        self.plotting = plotting
        self.coordinate_type = None
        self.require_variance = require_variance
        self.variance = None
    	
        if coordinate_type == "Eucledian":
            self.coordinate_type = 'euclidean'
        else:
            self.coordinate_type = 'geographic'

        self.nlags = nlags


    def _fit(self, X, y):
        """This method of the Kriging Class is used to fit Kriging interpolation model to
        the train data. This function shouldn't be called directly."""
        if self.type == "Ordinary":
            self.ok = OrdinaryKriging(
                X[:,0],
                X[:,1],
                y,
                variogram_model = self.variogram_model,
                enable_plotting = self.plotting,
                coordinates_type = self.coordinate_type,
                nlags = self.nlags
            )

        elif self.type == "Universal":
            self.uk = UniversalKriging(
                X[:,0],
                X[:,1],
                y,
                variogram_model = self.variogram_model,
                enable_plotting = self.plotting,
            )

        else:
            raise ValueError("Choose either Universal or Ordinary - Given argument is neither")


        return self

    def _predict_grid(self, x1lim, x2lim):
        """The function that is called to return the interpolated data in Kriging Interpolation
        in a grid. This method shouldn't be called directly"""
        lims = (*x1lim, *x2lim)
        x1min, x1max, x2min, x2max = lims
        x1 = np.linspace(x1min, x1max, self.resolution)
        x2 = np.linspace(x2min, x2max, self.resolution)

        if self.ok is not None:
            predictions, self.variance = self.ok.execute(
            style = 'grid',
            xpoints = x1,
            ypoints = x2
            )
        
        else:
            predictions, self.variance = self.uk.execute(
            style = 'grid',
            xpoints = x1,
            ypoints = x2
            )
            
        return  predictions

    def _predict(self, X):
        """This function should be called to return the interpolated data in kriging
        in a pointwise manner. This method shouldn't be called directly."""
        if self.ok is not None:
            predictions, self.variance = self.ok.execute(
            style = 'points',
            xpoints = X[:,0],
            ypoints = X[:,1]
            )
        
        else:
            predictions, self.variance = self.uk.execute(
            style = 'points',
            xpoints = X[:,0],
            ypoints = X[:,1]
            )
            
        return  predictions

    def return_variance(self):
        """This method of the Kriging class returns the variance at the interpolated
        points if the user chooses to use this option at the beginning of the interpolation"""
        if self.require_variance:
            return self.variance

        else:
            print("Variance not asked for, while instantiating the object. Returning None")
            return None




















"""
A module to have different distance metrics for spatial interpolation
"""
import numpy as np
from scipy.spatial.distance import cdist


def haversine(X1, X2):
    """
    Arguments
    ---------
    One test point
    Multiple Train Points

    Long Lat Order
    """

    # Non-vectorized version
    # X1 = X1.reshape(1, 2)
    # difference = (X1 - X2) * np.pi / 180
    # test_point_lat = X1[:, 1] * np.pi / 180
    # training_locations_lat = X2[:, 1] * np.pi / 180

    # a = np.sin(difference[:, 0] / 2)**2 * np.cos(test_point_lat) * np.cos(training_locations_lat) +\
    #     np.sin(difference[:, 1] / 2)**2
    # radius = 6371
    # c = 2 * np.arcsin(np.sqrt(a))
    # return radius * c

    # Vectorized code
    lon1, lat1, lon2, lat2 = map(
        np.radians, [X1[:, 0, None], X1[:, 1, None], X2[:, 0, None], X2[:, 1, None]])

    dlon = lon2.T - lon1
    dlat = lat2.T - lat1

    a = np.sin(dlat/2.0)**2 + np.cos(lat1)@np.cos(lat2.T) * np.sin(dlon/2.0)**2

    c = 2 * np.arcsin(np.sqrt(a))
    km = 6371 * c
    return km


def euclidean(X1, X2):
    # return np.linalg.norm(X1 - X2, 2, axis=1)
    return cdist(X1, X2)



import numpy as np
from scipy.optimize import curve_fit

from ..base import Base
from .polynomials import _create_polynomial


class Trend(Base):
    """Class to interpolate by fitting a curve to the data points
    available using `scipy`'s `curve_fit`.
    
    Parameters
    ----------
    order: int, default 1
        Selects the order of the polynomial to best fit.
        Possible values 0 <= order <= 2.

    custom_poly: functor, default None
        If you would like to fit to your custom function,
        _set order to None_ and then pass a functor.
        See Example functor passing below
        .. highlight:: python
        .. code-block:: python
            def func(X, a, b, c):
                x1, x2 = X
                return np.log(a) + b*np.log(x1) + c*np.log(x2)
            t = Trend(order=None, custom_poly=func)
            ...
    """

    def __init__(
        self,
        order=1,
        custom_poly=None,
        resolution="standard",
        coordinate_type="Euclidean",
    ):
        super().__init__(resolution, coordinate_type)
        self.order = order
        # setting the polynomial to fit our data
        if _create_polynomial(order) is not None:
            self.func = _create_polynomial(order)
        else:
            if custom_poly is not None:
                self.func = custom_poly
            else:
                raise ValueError("Arguments passed are not valid")

    def _fit(self, X, y):
        """Function for fitting trend interpolation.
        This function is not supposed to be called directly.
        """
        # fitting the curve using scipy
        self.popt, self.pcov = curve_fit(self.func, (X[:, 0], X[:, 1]), y)
        return self

    def _predict_grid(self, x1lim, x2lim):
        """Function for trend interpolation.
        This function is not supposed to be called directly.
        """
        # getting the boundaries for interpolation
        x1min, x1max = x1lim
        x2min, x2max = x2lim

        # forming the grid
        x1 = np.linspace(x1min, x1max, self.resolution)
        x2 = np.linspace(x2min, x2max, self.resolution)
        X1, X2 = np.meshgrid(x1, x2)
        return self.func((X1, X2), *self.popt)

    def _predict(self, X):
        """Function for random interpolation.
        This function is not supposed to be called directly.
        """
        x1, x2 = X[:, 0], X[:, 1]
        return self.func((x1, x2), *self.popt)



"""
This is a module for Natural Neighbors Interpolation
"""

import numpy as np
from scipy.spatial import Voronoi, voronoi_plot_2d
import matplotlib.pyplot as plt
from ..base import Base
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon
from math import atan2
from copy import deepcopy


def is_row_in_array(row, arr):
    return list(row) in arr.tolist()


def get_index(row, arr):
    t1 = np.where(arr[:, 0] == row[0])
    t2 = np.where(arr[:, 1] == row[1])
    index = np.intersect1d(t1, t2)[0]
    # If length of index exceeds one!! - Uniqueness Error
    return index


def order_poly(vertices):
    """This function essentially is used to order the vertices
    of the Voronoi polygon in a clockwise manner. This ensures
    that Shapely doesn't produce Polygon objects that are potentially
    non-convex and non-zero area.

    Arguments
    ---------
    vertices : {array-like, 2D matrix} 
    This contains the list of vertices of the Polygon to be sorted

    Returns
    -------
    new_vertices : {array-like, 2D matrix}
    All the vertices reordered in a clockwise manner
    """
    mean_x = np.mean(vertices[:, 0])
    mean_y = np.mean(vertices[:, 1])

    def condition(x):
        """This is the condition to be used while sorting. We convert the coordinates
        to Polar and sort the points
        """
        return atan2(x[0] - mean_x, x[1] - mean_y)*180/np.pi
    return sorted(vertices, key=condition)


class NaturalNeighbor(Base):
    """Class used for natural neighbors interpolation. This method is an implementation first
    proposed by Sibson et al. [1] in 1981. We use the weights derived using the work in [1]
    and leave it for future addition, the use of Laplace Weights [2].

    Parameters
    ----------
    weights: str, optional
        This defines the type of weights to be used for natural neighbor interpolation.
        We use Sibson Weights, and plan to add Laplace weights in the future
        Default value is "sibson"

    display: Boolean, optional
        True value displays the voronoi tesselation to the user after fitting the model.
        Default value is False.

    Notes 
    -----
    This is for contributors:
        The way in which part of the code is used is in the assumption that
        we use the data's ordering to find its voronoi partitions. 

    References
    ----------
    [1]  Sibson, R. (1981). "A brief description of natural neighbor interpolation (Chapter 2)". In V. Barnett (ed.). Interpolating Multivariate Data. Chichester: John Wiley. pp. 21–36.
    [2]  V.V. Belikov; V.D. Ivanov; V.K. Kontorovich; S.A. Korytnik; A.Y. Semenov (1997). "The non-Sibsonian interpolation: A new method of interpolation of the values of a function on an arbitrary set of points". Computational mathematics and mathematical physics. 37 (1): 9–15.
    [3]  N.H. Christ; R. Friedberg, R.; T.D. Lee (1982). "Weights of links and plaquettes in a random lattice". Nuclear Physics B. 210 (3): 337–346.
    """

    def __init__(
        self,
        weights="sibson",
        display=False,
        resolution="standard",
        coordinate_type="Eucledian",
    ):
        super().__init__(resolution, coordinate_type)
        self.weights = weights
        self.X = None
        self.y = None
        self.result = None
        self.voronoi = None
        self.vertices = None  # This variable stored the voronoi partition's vertices
        self.vertex_poly_map = dict()  # This variable stores the polygon to data point map
        self.display = display

    def _fit(self, X, y):
        """This function is for the natural neighbors interpolation method.
        This is not expected to be called directly.
        """
        self.X = X
        self.y = y
        self.voronoi = Voronoi(X, incremental=True)
        self.vertices = self.voronoi.vertices

        self.vertex_poly_map = {i: 0 for i in range(len(X))}

        for i in range(len(self.X)):
            index = np.where(self.voronoi.point_region == i)[0][0]
            point = Point(self.X[index])
            region = self.voronoi.regions[i]
            if -1 not in region and region != []:
                # -1 corresponds to unbounded region - we can't have this in interpolation
                # and the function returns an empty list anyways
                # at least in the case of non-incremental NN
                p = Polygon(order_poly(self.vertices[region]))
                self.vertex_poly_map[index] = p
        # Remove all the data points that do not contribute to Nearest Neighhbor interpolation
        for i in range(len(self.vertex_poly_map)):
            if self.vertex_poly_map[i] == 0:
                self.vertex_poly_map.pop(i, None)

        if self.display:
            voronoi_plot_2d(self.voronoi)
            plt.show()
            self.display = False

        return self

    def _predict_grid(self, x1lim, x2lim):
        """ Gridded interpolation for natural neighbors interpolation. This function should not
        be called directly. 
        """
        lims = (*x1lim, *x2lim)
        x1min, x1max, x2min, x2max = lims
        x1 = np.linspace(x1min, x1max, self.resolution)
        x2 = np.linspace(x2min, x2max, self.resolution)
        X1, X2 = np.meshgrid(x1, x2)
        return self._predict(np.array([X1.ravel(), X2.ravel()]).T)

    def _predict(self, X):
        """The function taht is called to predict the interpolated data in Natural Neighbors
        interpolation. This should not be called directly. 
        If this method returns None, then we cannot interpolate because of the formed Voronoi
        Tesselation
        """
        result = np.zeros(len(X))
        # Potentially create so many class objects as the
        # length of the to be predicted array
        # not a bad idea if memory is not a constraints
        for index in range(len(X)):

            if is_row_in_array(X[index], self.X):

                idx = get_index(X[index], self.X)
                # Check if query data point already exists
                result[index] = self.y[idx]

            else:
                # QHull object can't bgit ae pickled. Deepcopy doesn't work.
                # So we need to fit the model for each and every query data point.
                self._fit(self.X, self.y)

                vor = self.voronoi
                vor.add_points(np.array([X[index]]))
                vor.close()
                # We exploit the incremental processing of Scipy's Voronoi.
                # We create a copy to ensure that the original copy is preserved.
                new_regions = vor.regions
                new_vertices = vor.vertices
                final_regions = []

                for i in new_regions:
                    if i != [] and -1 not in i:
                        final_regions.append(i)

                new = []  # this stores the newly created voronoi partitions
                for i in range(len(new_vertices)):
                    if new_vertices[i] not in self.vertices:
                        new.append(new_vertices[i])
                new = np.array(new)
                if len(new) < 3:
                    # We need atleast a traingle to interpolate
                    # Three new voronoi vertices form a triangle
                    result[index] = np.nan
                    continue

                weights = {}  # Weights that we use for interpolation
                new_polygon = Polygon(order_poly(new))
                new_polygon_area = new_polygon.area

                for i in self.vertex_poly_map:
                    if new_polygon.intersects(self.vertex_poly_map[i]):
                        weights[i] = (new_polygon.intersection(
                            self.vertex_poly_map[i])).area/new_polygon_area

                prediction = np.array([self.y[i]*weights[i]
                                      for i in weights]).sum()
                result[index] = prediction
                del vor, weights, new_polygon, new_polygon_area

        return result



import numpy as np

from ..base import Base


class CustomInterpolator(Base):
    """
    Class to interpolate by fitting a sklearn type Regressor to
    the given data.

    Parameters
    ----------
    regressor: class definition,
        This variable is used to pass in the Regressor we would like
        to use for interpolation. The regressor sould be sklearn type
        regressor. Example from sklearn.ensemble -> RandomForestRegressor

    reg_kwargs: dict, optional
        This is a dictionary that is passed into the Regressor initialization.
        Use this to change the behaviour of the passed regressor. Default = empty dict

    Attributes
    ----------
    reg : object
        Object of the `regressor` class passed.
    """

    def __init__(
        self,
        regressor,
        resolution="standard",
        coordinate_type="Euclidean"
    ):
        super().__init__(resolution, coordinate_type)
        self.reg = regressor

    def _fit(self, X, y):
        """Function for fitting.
        This function is not supposed to be called directly.
        """
        self.reg.fit(X, y)
        return self

    def _predict_grid(self, x1lim, x2lim):
        """Function for grid interpolation.
        This function is not supposed to be called directly.
        """
        # getting the boundaries for interpolation
        x1min, x1max = x1lim
        x2min, x2max = x2lim

        # building the grid
        x1 = np.linspace(x1min, x1max, self.resolution)
        x2 = np.linspace(x2min, x2max, self.resolution)
        X1, X2 = np.meshgrid(x1, x2)
        return self.reg.predict(np.asarray([X1.ravel(), X2.ravel()]).T)

    def _predict(self, X):
        """Function for interpolation on specific points.
        This function is not supposed to be called directly.
        """
        return self.reg.predict(X)

    def __repr__(self):
        return self.__class__.__name__ + "." + self.reg.__class__.__name__



from .random.random import Random
from .idw.idw import IDW
from .spline.bspline import Spline
from .trend.trend import Trend
from .spatial.spatial import SpatialAverage
from .natural_neighbors.natural_neighbors import NaturalNeighbor
from .kriging.kriging import Kriging
from .gp.gp import GP
from .custom.custom import CustomInterpolator



from ..constants import RESOLUTION


class Base:
    """A class that is declared for performing Interpolation.
    This class should not be called directly, use one of it's
    children.
    """

    def __init__(self, resolution='standard', coordinate_types='Euclidean'):
        self.resolution = RESOLUTION[resolution]
        self.coordinate_type = coordinate_types
        self._fit_called = False

    def fit(self, X, y, **kwargs):
        """The function call to fit the model on the given data.

        Parameters
        ----------

        X: {array-like, 2D matrix}, shape(n_samples, 2)
            The set of all coordinates, where we have ground truth
            values
        y: array-like, shape(n_samples,)
            The set of all the ground truth values using which
            we perform interpolation

        Returns
        -------

        self : object
            Returns self

        """
        assert len(X.shape) == 2, "X must be a 2D array got shape = "+str(X.shape)
        # assert X.shape[1] == 2, "X can not have more than 2 dimensions"
        assert len(y.shape) == 1, "y should be a 1d array"
        assert y.shape[0] == X.shape[0], "X and y must be of the same size"

        # saving that fit was called
        self._fit_called = True

        # saving boundaries
        self.x1min_d = min(X[:, 0])
        self.x1max_d = max(X[:, 0])
        self.x2min_d = min(X[:, 1])
        self.x2max_d = max(X[:, 1])
        return self._fit(X, y, **kwargs)  # calling child specific fit method

    def predict(self, X, **kwargs):
        """The function call to return interpolated data on specific
        points.

        Parameters
        ----------

        X: {array-like, 2D matrix}, shape(n_samples, 2)
            The set of all coordinates, where we have ground truth
            values

        Returns
        -------

        y_pred : array-like, shape(n_samples,)
            The set of interpolated values for the points used to
            call the function.
        """

        assert len(X.shape) == 2, "X must be a 2D array got shape = "+str(X.shape)
        # assert X.shape[1] == 2, "X can not have more than 2 dimensions"

        # checking if model is fitted or not
        assert self._fit_called, "First call fit method to fit the model"

        # calling child specific _predict method
        return self._predict(X, **kwargs)

    def predict_grid(self, x1lim=None, x2lim=None, support_extrapolation=True):
        """Function to interpolate data on a grid of given size.
        .
        Parameters
        ----------
        x1lim: tuple(float, float),
            Upper and lower bound on 1st dimension for the interpolation.

        x2lim: tuple(float, float),
            Upper and lower bound on 2nd dimension for the interpolation.

        Returns
        -------
        y: array-like, shape(n_samples,)
            Interpolated values on the grid requested.
        """
        # checking if model is fitted or not
        assert self._fit_called, "First call fit method to fit the model"

        # by default we interpolate over the whole grid
        if x1lim is None:
            x1lim = (self.x1min_d, self.x1max_d)
        if x2lim is None:
            x2lim = (self.x2min_d, self.x2max_d)
        (x1min, x1max) = x1lim
        (x2min, x2max) = x2lim

        # extrapolation isn't supported yet
        if not support_extrapolation:
            assert self.x1min_d >= x1min, "Extrapolation not supported"
            assert self.x1max_d <= x1max, "Extrapolation not supported"
            assert self.x2min_d >= x2min, "Extrapolation not supported"
            assert self.x2max_d <= x2max, "Extrapolation not supported"

        # calling child specific _predict_grid method
        pred_y = self._predict_grid(x1lim, x2lim)
        return pred_y.reshape(self.resolution, self.resolution)

    def __repr__(self):
        return self.__class__.__name__

    def _fit(self, X, y):
        raise NotImplementedError

    def _predict_grid(self, x1lim, x2lim):
        raise NotImplementedError

    def _predict(self, X):
        raise NotImplementedError



"""This python script contains all the constants that
might be needed in the various interpolation pacakages. 
"""

low_res = 10
med_res = 100
high_res = 1000

RESOLUTION = {"low": low_res, "standard": med_res, "high": high_res}



"""
This is a module for GP Interpolation
"""
import numpy as np
from ..base import Base
from GPy.models import GPRegression
from GPy.kern import RBF


class GP(Base):
    """A class that is declared for performing GP interpolation.
    GP interpolation (usually) works on the principle of finding the 
    best unbiased predictor. 

    Parameters
    ----------
    type : str, optional
    This parameter defines the type of Kriging under consideration. This 
    implementation uses PyKrige package  (https://github.com/bsmurphy/PyKrige).
    The user needs to choose between "Ordinary" and "Universal".

    """

    def __init__(
        self,
        kernel=RBF(2, ARD=True),
    ):

        super().__init__()
        self.kernel = kernel

    def _fit(self, X, y, n_restarts=5, verbose=False, random_state=None):
        """ Fit method for GP Interpolation
        This function shouldn't be called directly.
        """
        np.random.seed(random_state)
        if len(y.shape) == 1:
            y = y.reshape(-1, 1)
        self.model = GPRegression(X, y, self.kernel)
        self.model.optimize_restarts(n_restarts, verbose=verbose)
        return self

    def _predict_grid(self, x1lim, x2lim):
        """The function that is called to return the interpolated data in Kriging Interpolation
        in a grid. This method shouldn't be called directly"""
        lims = (*x1lim, *x2lim)
        x1min, x1max, x2min, x2max = lims
        x1 = np.linspace(x1min, x1max, self.resolution)
        x2 = np.linspace(x2min, x2max, self.resolution)

        X1, X2 = np.meshgrid(x1, x2)
        X = np.array([(i, j) for i, j in zip(X1.ravel(), X2.ravel())])

        predictions = self.model.predict(X)[0].reshape(len(x1), len(x2))

        return predictions.ravel()

    def _predict(self, X, return_variance=False):
        """This function should be called to return the interpolated data in kriging
        in a pointwise manner. This method shouldn't be called directly."""

        predictions, variance = self.model.predict(X)
        if return_variance:
            return predictions.ravel(), variance
        else:
            return predictions.ravel()



"""
This is a module for inverse distance weighting (IDW) Spatial Interpolation
"""
import numpy as np
from ..utils.distance import haversine, euclidean
from ..base import Base


class IDW(Base):
    """A class that is declared for performing IDW Interpolation.
    For more information on how this method works, kindly refer to
    https://en.wikipedia.org/wiki/Inverse_distance_weighting

    Parameters
    ----------
    exponent : positive float, optional
        The rate of fall of values from source data points.
        Higher the exponent, lower is the value when we move
        across space. Default value is 2.

    Attributes
    ----------
    Interpolated Values : {array-like, 2D matrix}, shape(resolution, resolution)
    This contains all the interpolated values when the interpolation is performed
    over a grid, instead of interpolation over a set of points.

    X : {array-like, 2D matrix}, shape(n_samples, 2)
    Set of all the coordinates available for interpolation.

    y : array-like, shape(n_samples,)
    Set of all the available values at the specified X coordinates.

    result : array_like, shape(n_to_predict, )
    Set of all the interpolated values when interpolating over a given
    set of data points.

    """

    def __init__(self, exponent=2, resolution="standard", coordinate_type="Euclidean"):
        super().__init__(resolution, coordinate_type)
        self.exponent = exponent
        self.interpolated_values = None
        self.X = None
        self.y = None
        self.result = None
        if self.coordinate_type == 'Geographic':
            self.distance = haversine
        elif self.coordinate_type == 'Euclidean':
            self.distance = euclidean
        else:
            raise NotImplementedError(
                "Only Geographic and Euclidean Coordinates are available")

    def _fit(self, X, y):
        """This function is for the IDW Class.
        This is not expected to be called directly
        """
        self.X = X
        self.y = y
        return self

    def _predict_grid(self, x1lim, x2lim):
        """ Gridded interpolation for natural neighbors interpolation. This function should not
        be called directly. 
        """
        lims = (*x1lim, *x2lim)
        x1min, x1max, x2min, x2max = lims
        x1 = np.linspace(x1min, x1max, self.resolution)
        x2 = np.linspace(x2min, x2max, self.resolution)
        X1, X2 = np.meshgrid(x1, x2)
        return self._predict(np.array([X1.ravel(), X2.ravel()]).T)

    def _predict(self, X):
        """The function call to predict using the interpolated data
        in IDW interpolation. This should not be called directly.
        """

        dist = self.distance(self.X, X)
        weights = 1 / np.power(dist, self.exponent)
        result = (weights * self.y[:, None]).sum(axis=0) / weights.sum(axis=0)

        # if point is from train data, ground truth must not change
        for i in range(X.shape[0]):
            mask = np.equal(X[i], self.X).all(axis=1)
            if mask.any():
                result[i] = (self.y*mask).sum()

        return result



"""File containing polynomials supported by Trend interpolation
Class.
"""


def _create_polynomial(order):
    if order is None:  # custom function by the user
        return None

    elif order == 0:

        def func(X, a):
            return a

    elif order == 1:

        def func(X, a, b, c):
            x1, x2 = X
            return a + b * x1 + c * x2

    elif order == 2:

        def func(X, a, b, c, d, e, f):
            x1, x2 = X
            return a + b * x1 + c * x2 + d * (x1 ** 2) + e * (x1 ** 2) + f * x1 * x2

    else:
        raise NotImplementedError(
            f"{order} order polynomial needs to be defined manually"
        )

    return func



import numpy as np
from scipy.interpolate import bisplrep, bisplev


from ..base import Base
from ..utils import find_closest


class Spline(Base):
    """
    Class to use a bivariate B-spline to interpolate values.
    https://docs.scipy.org/doc/scipy-0.18.1/reference/generated/scipy.interpolate.bisplrep.html#scipy.interpolate.bisplrep

    Parameters
    ----------
    kx, ky: int, int, optional
        The degrees of the spline (1 <= kx, ky <= 5). 
        Third order (kx=ky=3) is recommended.

    s : float, optional
        A non-negative smoothing factor. If weights correspond
        to the inverse of the standard-deviation of the errors
        in z, then a good s-value should be found in the
        range `(m-sqrt(2*m),m+sqrt(2*m))` where `m=len(x)`.
    """

    def __init__(
        self, kx=3, ky=3, s=None, resolution="standard", coordinate_type="Euclidean"
    ):
        super().__init__(resolution, coordinate_type)
        self.kx = kx
        self.ky = ky
        self.s = s

    def _fit(self, X, y):
        """The function call to fit the spline model on the given data.
        This function is not supposed to be called directly.        
        """
        # fitting the curve
        # bisplrep returns details of the fitted curve
        # read bisplrep docs for more info about it's return values.
        self.tck = bisplrep(X[:, 0], X[:, 1], y,
                            kx=self.kx, ky=self.ky, s=self.s)
        return self

    def _predict_grid(self, x1lim, x2lim):
        """The function to predict grid interpolation using the BSpline.
        This function is not supposed to be called directly.
        """
        # getting the boundaries for interpolation
        x1min, x1max = x1lim
        x2min, x2max = x2lim

        # interpolating over the grid
        # TODO Relook here, we might expect the result to be transpose
        return bisplev(
            np.linspace(x1min, x1max, self.resolution),
            np.linspace(x2min, x2max, self.resolution),
            self.tck,
        )

    def _predict(self, X):
        """The function to predict using the BSpline interpolation.
        This function is not supposed to be called directly.
        """
        results = []
        for ix in range(X.shape[0]):
            interpolated_y = bisplev(
                X[ix, 0], X[ix, 1], self.tck
            ).item()  # one value returned
            results.append(interpolated_y)

        return np.array(results)

        # # form a grid
        # x1 = np.linspace(self.x1min_d, self.x1max_d, self.resolution),
        # x2 = np.linspace(self.x2min_d, self.x2max_d, self.resolution),
        # X1, X2 = np.meshgrid(x1, x2)

        # # be default run grid interpolation on the whole train data
        # interpolated_grid = bisplev(
        #     x1, x2,
        #     self.tck,
        # )

        # # find the closest points on the interpolated grid
        # ix = find_closest(grid=(X1, X2), X)
        # return interpolated_grid[ix] # TODO this can be wrong, must depend on



import pytest
import numpy as np
from time import time

X = np.random.rand(20, 2)
y = np.random.rand(20)

X_new = np.random.rand(40, 2)


def common(model):
    print(repr(model), end=' ')
    init = time()
    model.fit(X, y)
    y_new = model.predict(X_new)

    assert y_new.shape == (40, )
    print('Passed', 'Time:', np.round(time()-init, 3), 'seconds')


def test_basic():
    from polire import (IDW, Spline, Trend, GP, Kriging,
                        NaturalNeighbor, SpatialAverage,
                        CustomInterpolator)
    from sklearn.linear_model import LinearRegression

    common(IDW())
    common(Spline())
    common(Trend())
    common(GP())
    common(Kriging())
    common(NaturalNeighbor())
    common(SpatialAverage())
    common(CustomInterpolator(LinearRegression(normalize=True)))


if __name__ == '__main__':
    test_basic()



import numpy as np

from ..base import Base


class Random(Base):
    """
    Class to randomly interpolate by picking values between maximum and
    minimum measurements.

    Note: Even if a point on the requested grid is present in 
    the training set, we return a random value for it.
    """

    def __init__(self, resolution="standard", coordinate_type="Euclidean"):
        super().__init__(resolution, coordinate_type)

    def _fit(self, X, y):
        """Function for fitting random interpolation.
        This function is not supposed to be called directly.
        """
        self.ymax = max(y)
        self.ymin = min(y)
        return self

    def _predict_grid(self, x1lim, x2lim):
        """Function for random grid interpolation.
        This function is not supposed to be called directly.
        """
        return np.random.uniform(
            low=self.ymin, high=self.ymax, size=(self.resolution, self.resolution)
        )

    def _predict(self, X):
        """Function for random interpolation.
        This function is not supposed to be called directly.
        """
        return np.random.uniform(low=self.ymin, high=self.ymax, size=(X.shape[0]))



""" Standard Utility Script for Gridding Data
    1. Contains all the common functions that 
        will be employed across various different interpolators

"""
import numpy as np
from scipy import spatial


def make_grid(self, x, y, res, offset=0.2):
    """ This function returns the grid to perform interpolation on.
       This function is used inside the fit() attribute of the idw class.
    
    Parameters
    ----------
    x: array-like, shape(n_samples,)
        The first coordinate values of all points where
        ground truth is available
    y: array-like, shape(n_samples,)
        The second coordinate values of all points where
        ground truth is available
    res: int
        The resolution value
    offset: float, optional
        A value between 0 and 0.5 that specifies the extra interpolation to be done
        Default is 0.2
    
    Returns
    -------
    xx : {array-like, 2D}, shape (n_samples, n_samples)
    yy : {array-like, 2D}, shape (n_samples, n_samples)
    """
    y_min = y.min() - offset
    y_max = y.max() + offset
    x_min = x.min() - offset
    x_max = x.max() + offset
    x_arr = np.linspace(x_min, x_max, res)
    y_arr = np.linspace(y_min, y_max, res)
    xx, yy = np.meshgrid(x_arr, y_arr)
    return xx, yy


def find_closest(grid, X, l=2):
    """Function used to find the indices of the grid points closest
    to the passed points in X.

    Parameters
    ----------
    grid: {list of 2 arrays}, (shape(res, res), shape(res, res))
        This is generated by meshgrid.

    X: {array-like, 2D matrix}, shape(n_samples, 2)
        The set of points to which we need to provide closest points
        on the grid.

    l: str, optional
        To decide the `l`th norm to use. `Default = 2`.

    Returns
    -------
    ix: array, shape(X.shape[0],)
        The index of the point closest to points in X.

    ref - https://stackoverflow.com/questions/10818546/finding-index-of-nearest-point-in-numpy-arrays-of-x-and-y-coordinates
    """
    points = np.asarray([grid[0].ravel(), grid[1].ravel()]).T  # ravel is inplace
    kdtree = spatial.KDTree(points)
    ixs = []  # for containing the indices of closest points found on grid

    for point_ix in range(X.shape[0]):
        point = X[point_ix, :]
        _, ix = kdtree.query(point)
        ixs.append(ix)

    return ixs



import numpy as np
from ..idw.idw import IDW
from ..utils.distance import haversine, euclidean


class SpatialFeatures:
    """Generate spatial features from N-closest locations

    Args:
        n_closest : 'N' closest locations

        idw : To use idw output as one of the feature

        idw_exponent : Exponent to be used in idw (if idw is False, ignore)

        coordinate_type : 'Eucleadian' or 'Geographic' (if idw is False, ignore)

        resolution : 'low', 'standard' or 'high' (if idw is False, ignore)
    """

    def __init__(self, n_closest: int = 5, idw: bool = True, idw_exponent: float = 2,
                 coordinate_type: str = 'Euclidean', resolution: str = 'standard') -> None:

        self.n_closest = n_closest
        self.idw = idw
        self.idw_exponent = idw_exponent
        self.coordinate_type = coordinate_type
        self.resolution = resolution
        if self.coordinate_type == 'Eucledian':
            self.distance = euclidean
        elif self.coordinate_type == 'Geographic':
            self.distance = haversine
        else:
            raise NotImplementedError(
                '"'+self.coordinate_type+'" is not implemented yet or invalid')

    def fit(self, X: np.ndarray, y: np.ndarray) -> object:
        """[summary]

        Args:
            X : Reference X data (longitude, latitude, time, ...)
            y : Reference y data

        Returns:
            self
        """
        self.X = X
        self.y = y

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Transform features

        Args:
            X (np.ndarray): (longitude, latitude, time, ...)

        Raises:
            Exception: If not already fitted

        Returns:
            np.ndarray: Transformed features
        """
        try:
            self.X
        except AttributeError:
            raise Exception("Not fitted yet. first call the 'fit' method")

        Xflag = False
        if np.all(X == self.X):
            Xflag = True

        F = np.empty((X.shape[0], (X.shape[1] - 3) +
                     self.n_closest*2 + self.idw)) * np.nan
        for t in np.unique(X[:, 2]):            # Iterating over time
            mask = X[:, 2] == t                 # rows with time t
            trn_mask = self.X[:, 2] == t
            X_local = X[mask]
            self_X_local = self.X[trn_mask]

            lonlat = X_local[:, :2]             # locs
            self_lonlat = self_X_local[:, :2]   # Reference locs
            dst = self.distance(lonlat, self_lonlat)
            if Xflag:
                idx = dst.argsort()[:, 1:self.n_closest+1]
            else:
                idx = dst.argsort()[:, :self.n_closest]

            # Feature set 1: closest distances
            f1 = dst[np.arange(lonlat.shape[0])[:, None], idx]

            self_y_local = self.y[trn_mask]     # Train obs
            ymat = self_y_local[:, None].repeat(lonlat.shape[0], 1).T
            # Feature set 2: closest observations
            f2 = ymat[np.arange(lonlat.shape[0])[:, None], idx]

            if self.idw:
                def for_each_row(i):
                    i = i[0]
                    model = IDW(exponent=self.idw_exponent)
                    model.resolution = self.resolution
                    model.coordinate_type = self.coordinate_type
                    model.fit(self_lonlat[idx[i]], self_y_local[idx[i]])
                    return model.predict(lonlat[i][None, :])

                # Feature set 3: IDW observation
                f3 = np.apply_along_axis(
                    for_each_row, axis=1, arr=np.arange(lonlat.shape[0]).reshape(-1, 1))
                F[mask] = np.concatenate([X_local[:, 3:], f1, f2, f3], axis=1)
            else:
                F[mask] = np.concatenate([X_local[:, 3:], f1, f2], axis=1)

        return F

    def fit_transform(self, X: np.ndarray, y: np.ndarray):
        self.fit(X, y)
        return self.transform(X)
